@@ -6,10 +6,33 @@ import pytest
 import itertools as it
 
 from pycvodes import (
-    integrate_adaptive, integrate_predefined, requires_jac, get_include
+    integrate_adaptive, integrate_predefined, requires_jac, get_include, iterative_linsols
 )
 
 
+def test_config_env():
+    # This test is useful for e.g. out-of-tree conda-recipes (e.g. conda-forge) where variables
+    # are set manually (and could otherwise be forgotten unless tests fail).
+    from .._config import env
+    from .._types import type_of_prec
+    assert env['REAL_TYPE'] in type_of_prec.values()
+    assert env['SUNDIALS_PRECISION'] in type_of_prec.keys()
+    for k in 'LAPACK SUNDIALS_LIBS INDEX_TYPE'.split():
+        assert k in env
+
+
+def requires_klu(_test):
+    from .._config import env
+    return pytest.mark.skipif(env.get('NO_KLU', '0') == '1',
+                              reason="sparse jacobian tests require KLU to be enabled")(_test)
+
+
+def high_precision(_test):
+    from .._config import env
+    return pytest.mark.skipif(env['SUNDIALS_PRECISION'] == "single",
+                              reason="test atol and/or rtol is too high for single precision")(_test)
+
+
 def test_get_include():
     assert get_include().endswith('include')
     assert 'cvodes_cxx.hpp' in os.listdir(get_include())
@@ -37,7 +60,7 @@ def decay_get_Cref(k, y0, tout):
             min(3, len(k)+1))])
 
 
-def _get_f_j(k):
+def _get_f_j(k, with_sparse=False):
     k0, k1, k2 = k
 
     def f(t, y, fout):
@@ -59,26 +82,19 @@ def _get_f_j(k):
             dfdx_out[0] = 0
             dfdx_out[1] = 0
             dfdx_out[2] = 0
-    return f, j
-
-
-def _get_f_j_sparse(k):
-    k0, k1, k2 = k
-
-    def f(t, y, fout):
-        fout[0] = -k0*y[0]
-        fout[1] = k0*y[0] - k1*y[1]
-        fout[2] = k1*y[1] - k2*y[2]
 
     _colptrs = [0, 2, 4, 5]
     _rowvals = [0, 1, 1, 2, 2]
 
-    def j(t, y, data, colptrs, rowvals):
+    def j_sparse(t, y, data, colptrs, rowvals):
         data[:] = -k0, k0, -k1, k1, -k2
         colptrs[:] = _colptrs
         rowvals[:] = _rowvals
 
-    return f, j, len(_rowvals)
+    if with_sparse:
+        return f, j, j_sparse, len(_rowvals)
+    else:
+        return f, j
 
 
 def _gravity_f_j_jtimes(g):
@@ -115,8 +131,6 @@ methods = [('adams', 1.8, False),
            ('bdf', 10, False),
            ('bdf', 10, True)]
 
-iterative_linear_solvers = ["gmres", "bicgstab", "tfqmr"]
-
 
 def bandify(cb, mlower, mupper):
     def j(t, y, jmat_out, dfdx_out=None, fy=None):
@@ -130,6 +144,8 @@ def bandify(cb, mlower, mupper):
 
 @pytest.mark.parametrize("method,forgiveness,banded", methods)
 def test_integrate_predefined(method, forgiveness, banded):
+    from .._config import env
+
     use_jac = method in requires_jac
     k = k0, k1, k2 = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
@@ -147,15 +163,20 @@ def test_integrate_predefined(method, forgiveness, banded):
 
     for j in jac_callbacks:
         xout = np.linspace(0, 3, 31)
-        atol, rtol = 1e-8, 1e-8
+        if env['SUNDIALS_PRECISION'] == "single":
+            atol = [1e-4, 3e-5, 2e-5]
+            rtol = 1e-4
+        else:
+            atol = [1e-8, 3e-9, 2e-9]
+            rtol = 1e-8
         # Run twice to catch possible side-effects:
-        yout, nfo = integrate_predefined(f, j, y0, xout, [1e-8, 3e-9, 2e-9], 1e-8, **kwargs)
-        yout, nfo = integrate_predefined(f, j, y0, xout, [1e-8, 3e-9, 2e-9], 1e-8, **kwargs)
+        yout, nfo = integrate_predefined(f, j, y0, xout, atol, rtol, **kwargs)
+        yout, nfo = integrate_predefined(f, j, y0, xout, atol, rtol, **kwargs)
         yref = decay_get_Cref(k, y0, xout)
         assert np.allclose(yout, yref,
                            rtol=forgiveness*rtol,
-                           atol=forgiveness*atol)
-        assert nfo['atol'] == [1e-8, 3e-9, 2e-9] and nfo['rtol'] == 1e-8
+                           atol=forgiveness*max(atol))
+        assert np.allclose(nfo['atol'], atol) and np.isclose(nfo['rtol'], rtol)
         assert nfo['nfev'] > 0
         if method in requires_jac and j is not None:
             assert nfo['njev'] > 0
@@ -163,6 +184,7 @@ def test_integrate_predefined(method, forgiveness, banded):
                 assert nfo['time_jac'] >= 0
 
 
+@high_precision
 def test_integrate_adaptive_tstop0():
     k = k0, k1, k2 = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
@@ -180,10 +202,15 @@ def test_integrate_adaptive_tstop0():
 
 @pytest.mark.parametrize("method,forgiveness,banded", methods)
 def test_integrate_adaptive(method, forgiveness, banded):
+    from .._config import env
+
     use_jac = method in requires_jac
     k = k0, k1, k2 = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
-    atol, rtol = 1e-8, 1e-8
+    if env['SUNDIALS_PRECISION'] == "single":
+        atol, rtol = 1e-4, 1e-4
+    else:
+        atol, rtol = 1e-8, 1e-8
     kwargs = dict(x0=0, xend=3, dx0=1e-10, atol=atol, rtol=rtol,
                   method=method, iter_type='newton')
     f, j = _get_f_j(k)
@@ -204,7 +231,7 @@ def test_integrate_adaptive(method, forgiveness, banded):
     assert info['nfev'] > 0
     if method in requires_jac:
         assert info['njev'] > 0
-    assert info['atol'] == [1e-8] and info['rtol'] == 1e-8
+    assert np.allclose(info['atol'], [atol]) and np.isclose(info['rtol'], rtol)
 
     with pytest.raises(RuntimeError) as excinfo:
         kw = kwargs.copy()
@@ -218,6 +245,7 @@ def test_integrate_adaptive(method, forgiveness, banded):
     assert '7' in str(excinfo.value).lower()
 
 
+@high_precision
 def test_derivative_1():
     def f(t, y, fout):
         fout[0] = y[0]
@@ -237,6 +265,7 @@ def test_derivative_1():
     assert '7' in str(excinfo.value).lower()
 
 
+@high_precision
 def test_derivative_2():
     def f(t, y, fout):
         fout[0] = y[0]
@@ -253,6 +282,7 @@ def test_derivative_2():
     assert np.allclose(yout, ref)
 
 
+@high_precision
 def test_derivative_3():
     def f(t, y, fout):
         fout[0] = y[1]
@@ -270,6 +300,7 @@ def test_derivative_3():
     assert np.allclose(discrepancy, 0, rtol=1e-6, atol=1e-6)
 
 
+@high_precision
 def test_roots_adaptive():
     def f(t, y, fout):
         fout[0] = y[0]
@@ -284,6 +315,7 @@ def test_roots_adaptive():
     assert np.min(np.abs(xout - 1)) < 1e-11
 
 
+@high_precision
 def test_quads_adaptive():
     k = 0.7
 
@@ -305,6 +337,7 @@ def test_quads_adaptive():
     assert np.allclose(info['quads'][:, 1], q1)
 
 
+@high_precision
 def test_quads_predefined():
     k = 0.7
 
@@ -327,6 +360,7 @@ def test_quads_predefined():
     assert np.allclose(info['quads'][:, 1], q1)
 
 
+@high_precision
 def test_roots_predefined():
     def f(t, y, fout):
         fout[0] = y[0]
@@ -343,6 +377,7 @@ def test_roots_predefined():
     assert info['root_indices'][0] == 2
 
 
+@high_precision
 def test_roots_predefined_autorestart():
     def f(t, y, fout):
         fout[0] = y[0]
@@ -379,6 +414,7 @@ def test_adaptive_nderiv():
     assert np.allclose(discrepancy, 0, atol=1e-3)
 
 
+@high_precision
 def test_return_on_root():
     def f(t, y, fout):
         fout[0] = y[0]
@@ -395,6 +431,7 @@ def test_return_on_root():
     assert info['success']
 
 
+@high_precision
 def test_predefined_roots_output():
     def f(t, y, fout):
         fout[0] = y[0]
@@ -423,6 +460,7 @@ def test_rhs_unrecoverable_error():
         yout, info = integrate_predefined(f, None, [1], [0, 1, 2], **kwargs)
 
 
+@high_precision
 def test_rhs_recoverable_error():
     global idx
     idx = -2
@@ -437,6 +475,7 @@ def test_rhs_recoverable_error():
     yout, info = integrate_predefined(f, None, [1], [0, 1, 2], **kwargs)
 
 
+@high_precision
 def test_adaptive_return_on_error():
     k = k0, k1, k2 = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
@@ -458,6 +497,7 @@ def test_adaptive_return_on_error():
     assert xout[-1] < kwargs['xend']  # obviously not strict
 
 
+@high_precision
 def test_adaptive_autorestart():
     k = k0, k1, k2 = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
@@ -479,6 +519,7 @@ def test_adaptive_autorestart():
     assert xout[-1] == kwargs['xend']
 
 
+@high_precision
 def test_predefined_autorestart():
     k = k0, k1, k2 = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
@@ -502,6 +543,7 @@ def test_predefined_autorestart():
     assert xout[-1] == xend
 
 
+@high_precision
 def test_predefined_return_on_error():
     k = 2.0, 3.0, 4.0
     y0 = [0.7, 0., 0.]
@@ -522,6 +564,7 @@ def test_predefined_return_on_error():
     assert info['success'] is False
 
 
+@high_precision
 def test_dx0cb():
     k = 1e23, 3.0, 4.0
     y0 = [.7, .0, .0]
@@ -538,6 +581,7 @@ def test_dx0cb():
     assert xout[-1] == xend
 
 
+@high_precision
 def test_dx_max_cb():
     k = 1e23, 3.0, 4.0
     y0 = [.7, .0, .0]
@@ -554,6 +598,7 @@ def test_dx_max_cb():
     assert xout[-1] == xend
 
 
+@high_precision
 def test_predefined_ew_ele():
     k = 2.0, 3.0, 4.0
     y0 = [0.7, 0., 0.]
@@ -573,6 +618,7 @@ def test_predefined_ew_ele():
     assert np.all(abs_ew_ele < 1)
 
 
+@high_precision
 def test_adaptive_ew_ele():
     k = 2.0, 3.0, 4.0
     y0 = [0.7, 0., 0.]
@@ -591,9 +637,10 @@ def test_adaptive_ew_ele():
     assert np.all(abs_ew_ele < 1)
 
 
-jtimes_params = list(it.product(iterative_linear_solvers, [True, False]))
+jtimes_params = list(it.product(iterative_linsols, [True, False]))
 
 
+@high_precision
 @pytest.mark.parametrize("linear_solver,with_jac", jtimes_params)
 def test_jtimes_adaptive(linear_solver, with_jac):
     g = 9.81
@@ -614,6 +661,7 @@ def test_jtimes_adaptive(linear_solver, with_jac):
         assert info['njev'] == 0
 
 
+@high_precision
 @pytest.mark.parametrize("linear_solver,with_jac", jtimes_params)
 def test_jtimes_predefined(linear_solver, with_jac):
     g = 9.81
@@ -635,38 +683,55 @@ def test_jtimes_predefined(linear_solver, with_jac):
         assert info['njev'] == 0
 
 
+@high_precision
+@requires_klu
+def test_sparse_equals_dense():
+    try:
+        from scipy.sparse import csc_matrix
+    except ImportError:
+        pytest.skip("Need scipy.sparse to compare sparse/dense jacobian output")
+    k = 2.0, 3.0, 4.0
+    y0 = [0.7, 0.3, 0.5]
+    f, j, j_sparse, nnz = _get_f_j(k, with_sparse=True)
+
+    n = len(y0)
+    J_dense = np.empty((n, n))
+    data = np.empty(nnz)
+    colptrs = np.empty(n + 1)
+    rowvals = np.empty(nnz)
+    j(0, y0, J_dense)
+    j_sparse(0, y0, data, colptrs, rowvals)
+
+    J_sparse = csc_matrix((data, rowvals, colptrs), shape=(n, n)).toarray()
+    assert np.allclose(J_sparse, J_dense)
+
+
+@high_precision
+@requires_klu
 def test_sparse_jac_adaptive():
-    from .._config import env
-    if env['LAPACK'] in ('', '0'):
-        pytest.skip("sparse jacobian tests require BLAS/LAPACK for KLU solver")
-    if env['NO_KLU'] == '1':
-        pytest.skip("sparse jacobian tests require KLU to be enabled (-DPYCVODES_NO_KLU=0)")
     k = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
-    f, j, nnz = _get_f_j_sparse(k)
     atol, rtol = 1e-8, 1e-8
+    f, _, j_sparse, nnz = _get_f_j(k, with_sparse=True)
     kwargs = dict(atol=atol, rtol=rtol, method='bdf',
                   linear_solver='klu', nnz=nnz)
-    xout, yout, info = integrate_adaptive(f, j, y0, 0, 10, **kwargs)
+    xout, yout, info = integrate_adaptive(f, j_sparse, y0, 0, 10, **kwargs)
     yref = decay_get_Cref(k, y0, xout - xout[0])
     assert info['success']
     assert info['njev'] > 0
     assert np.allclose(yout, yref, rtol=10*rtol, atol=10*atol)
 
 
+@high_precision
+@requires_klu
 def test_sparse_jac_predefined():
-    from .._config import env
-    if env['LAPACK'] in ('', '0'):
-        pytest.skip("sparse jacobian tests require BLAS/LAPACK for KLU solver")
-    if env['NO_KLU'] == '1':
-        pytest.skip("sparse jacobian tests require KLU to be enabled (-DPYCVODES_NO_KLU=0)")
     k = 2.0, 3.0, 4.0
     y0 = [0.7, 0.3, 0.5]
     xout = np.linspace(0, 3, 31)
-    f, j, nnz = _get_f_j_sparse(k)
     atol, rtol = 1e-8, 1e-8
+    f, _, j_sparse, nnz = _get_f_j(k, with_sparse=True)
     kwargs = dict(method='bdf', linear_solver='klu', nnz=nnz)
-    yout, info = integrate_predefined(f, j, y0, xout, 1.0e-8, 1.0e-8, **kwargs)
+    yout, info = integrate_predefined(f, j_sparse, y0, xout, 1.0e-8, 1.0e-8, **kwargs)
     yref = decay_get_Cref(k, y0, xout - xout[0])
     assert info['success']
     assert info['njev'] > 0
